#include "extras.lua"
#include "particles.lua"
#include "shake.lua"
#include "ufo.lua"
#include "projectile.lua"
#include "toolUIScripts/toolUIManager.lua"

function GetOrSetFloat(path, default)
    return ((HasKey(path)) and GetFloat(path, default) or SetFloat(path, default))
end
--Gets config values
function getConfig()
    return {
        theme = {
            color_r = GetOrSetFloat("savegame.mod.abductizer.custom.beam.color.r", 0),
            color_g = GetOrSetFloat("savegame.mod.abductizer.custom.beam.color.g", 1),
            color_b = GetOrSetFloat("savegame.mod.abductizer.custom.beam.color.b", 0),
        },
        controls = {
            action_1 = "usetool",
            action_2 = "grab",
            inspectMenu = "mmb",
        },
    }
end

config = getConfig()

--Called at the beginning
function init()
    RegisterTool("abductizer", "Abductizer", "MOD/vox/model.vox", 6)
    SetBool("game.tool.abductizer.enabled", true)

    --Sounds
    hoverSnd = LoadLoop("MOD/snd/ufo_hover.ogg")
    beamSnd = LoadLoop("MOD/snd/ufo_tractorbeam.ogg")
    motorSnd = LoadLoop("MOD/snd/motor.ogg")
    beamBegin = LoadSound("MOD/snd/ufo_beam_begin.ogg")
    beamEnd = LoadSound("MOD/snd/ufo_beam_end.ogg")
    destroySnd = LoadSound("MOD/snd/ufo_destroy.ogg")
    signalSnd = LoadSound("MOD/snd/signal.ogg")
    signalEndSnd = LoadSound("MOD/snd/signal_end.ogg")

    --Sprites
    beamSprite = LoadSprite("MOD/sprite/tractorbeam.png")
    signalSprite = LoadSprite("MOD/sprite/signal.png")
    ufoSprite = LoadSprite("MOD/sprite/ufo.png")
    arrowSprite = LoadSprite("MOD/sprite/arrow.png")
    glowSprite = LoadSprite("MOD/sprite/glow.png")
end
--Called every frame
function tick(dt)
    --Get new config values
    config = getConfig()

    tickUfo(dt)
    tickSignalEndAnimations(dt)
    tickToolUIManager()

    --If tool should work continue
    if GetString("game.player.tool") ~= "abductizer" then return end
    if not GetBool("game.player.canusetool") then return end

    tickTool(dt)

    --Process shaking if needed
    processShake(dt)
end
--Beam particles
function update(dt)
    updateSignals(dt)
    ufoUpdate(dt)
end
--Draws UI
function draw()
    drawUfoCallOffTimer()
end