#include "main.lua"

--Constant UFO values
ufoConsts = {}
    ufoConsts["tractorBeamDistance"] = 100
    ufoConsts["tractorBeamSize"] = 1.5
    ufoConsts["tractorBeamEffectSize"] = 4
    ufoConsts["tractorBeamDamping"] = 1.06
    ufoConsts["colSize"] = 5
    ufoConsts["hoverHeight"] = 14
    ufoConsts["voxelHungerGain"] = 0.08
    ufoConsts["voxelsToMinPower"] = 25000
    ufoConsts["voxelsToLeave"] = 200
    ufoConsts["timeUntilGiveUp"] = 6
    ufoConsts["timeToLeave"] = 4
    ufoConsts["beamDestructionTipBias"] = 1
    ufoConsts["staticThreshold"] = 1  
    ufoConsts["randOffsetSpeed"] = 2
    ufoConsts["randOffsetPower"] = 1
    ufoConsts["distancePowerCurve"] = {
        {point=0, value=0},
        {point=0.5, value=1},
        {point=1, value=0},
    }
ufo = {}

--Create dynamic UFO values
function createUfoProperties()
    --Dynamic UFO values, can be different for each UFO
    --These values are going to be adjusted later

    ufo["beamOn"] = false
    ufo["beamEnd"] = Vec(0, 0, 0)
    ufo["lastOffset"] = Vec(0, 0, 0)
    ufo["offsetPos"] = getBodyPos(ufo.handle)
    ufo["curPos"] = VecSub(getBodyPos(ufo.handle), ufo.lastOffset)
    ufo["destroyOnReachFlightEnd"] = false
    ufo["pullPower"] = 0
    ufo["tractorBeamPower"] = 0
    ufo["pullSize"] = 0
    ufo["maxAbductVoxels"] = 0
    glowPower = 0
    ufo["breakDistance"] = 0
    ufo["voxelsTakenByTime"] = ufoConsts.voxelsToMinPower
    ufo["giveUpTimer"] = ufoConsts.timeUntilGiveUp
    ufo["isLeaving"] = false
    ufo["flight"] = {}
        ufo["flight"]["startingPos"] = nil
        ufo["flight"]["targetPos"] = nil
        ufo["flight"]["time"] = nil
        ufo["flight"]["timePassed"] = nil
        ufo["flight"]["destroyOnReachFlightEnd"] = nil
        ufo["flight"]["easing"] = nil

    --UFO shapes
    ufo["beamEmitterShape"] = GetBodyShapes(ufo.handle)[5]
    ufo["beamFrameShape"] = GetBodyShapes(ufo.handle)[1]
end
function drawUfoCallOffTimer()
    --Draw UFO leave time
    if ufo.handle ~= nil and tool.callingUfoOff then
        UiPush()
            local x, y, dist = UiWorldToPixel(VecAdd(ufo.offsetPos, Vec(0, 1, 0)))
            UiTranslate(x, y)
            UiAlign("center middle")
            UiScale(math.max(20 / dist, 0))

            local split = 2
            local sprites = 8
            local progress = (tool.timeToCallUfoOffLeft - math.floor(tool.timeToCallUfoOffLeft*split)/split)*split

            --Create circle
            UiPush()
                local col = VecLerp(
                    Vec(config.theme.color_r-0.25, config.theme.color_g-0.25, config.theme.color_b-0.25),
                    Vec(1, 1, 1),
                    progress)
                UiColor(col[1], col[2], col[3])
                UiScale(lerp(0.5, 1, progress))
                UiImage("ui/world/circle/" .. 
                    tostring(
                        math.ceil(progress*sprites)
                    )
                    .. ".png")
            UiPop()

            --Create countdown
            UiTextShadow(0, 0, 0, 0.5, progress*5, 0)
            UiFont("bold.ttf", 84)
            UiText(tostring(math.ceil(tool.timeToCallUfoOffLeft*split)))
        UiPop()
    end
end
--Call a new UFO
function callNewUFO(callPos)
    --Create new UFO if one already doesn't exist  
    if ufo.handle == nil then         
        createUfo(callPos)
        createUfoProperties()
        SetShapeEmissiveScale(ufo.beamEmitterShape, 0)
    end   

    ufo.isLeaving = false

    local flyPos = VecAdd(callPos, Vec(0, ufoConsts.hoverHeight, 0))
        
    if VecDist(flyPos, ufo.curPos) > 40 then
        beamStateChange(false)
        planPath(ufo.curPos, flyPos, 2.5, false, easeInOutElastic)
    else
        planPath(ufo.curPos, flyPos, 4.5, false, easeOutQuart)            
    end
end
--Turn beam on or off
function beamStateChange(isOn)
    if ufo.beamOn == isOn then return end

    --Particles
    beamOnOffParticles(VecSub(ufo.offsetPos, Vec(0, 0.2, 0)))

    if isOn then
        PlaySound(beamBegin, ufo.curPos, 1.5)
        SetShapeEmissiveScale(ufo.beamEmitterShape, 10)

        --Flash glow
        glowPower = 500
        SetValue("glowPower", 65, "easeout", 0.5)

        --Shake
        shakeSet(1, 40, Vec(0, 1, 0), 0.35)
    else 
        PlaySound(beamEnd, ufo.curPos, 1.5)
        SetShapeEmissiveScale(ufo.beamEmitterShape, 0)
    end

    ufo.beamOn = isOn
end
--Set fly path from point A to point B
function planPath(from, to, speed, delete, easing)
    --Set new values
    ufo.flight.startingPos = ufo.curPos
    ufo.flight.targetPos = to
    ufo.flight.time = speed
    ufo.flight.timePassed = 0
    ufo.flight.destroyOnReachFlightEnd = delete
    ufo.flight.easing = easing

    --Reset voxel time
    ufo.voxelsTakenByTime = ufoConsts.voxelsToMinPower
end
--Makes UFO leave
function ufoLeave()
    ufo.isLeaving = true

    beamStateChange(false)

    planPath(ufo.curPos, VecAdd(ufo.curPos, Vec(0, 100, 0)), ufoConsts.timeToLeave, true, easeInQuart)
end
--Creates a new UFO
function createUfo(newPos)
    local newRandomPos = VecAdd(
        VecScale(
        --Make spawn points circular around the called point
        VecNormalize(
            Vec(math.random(-100, 100), 100, math.random(-100, 100))),
        100),
        newPos)
    --Create UFO
    ufo["handle"] =
        Spawn("MOD/vox/ufo/ufo.xml", 
        Transform(newRandomPos))[1]
end
--Progress UFO
function tickUfo(dt)
    if ufo.handle == nil then return end

    --Update position and offset
    local randOffset = updateUfoPosVariables()
    --Store last offset for the next iteration
    ufo.lastOffset = VecCopy(randOffset)

    --Set as non dynamic
    SetBodyDynamic(ufo.handle, false)
        
    --Rotates UFO beam frame around it's center
    rotateShape(ufo.beamFrameShape,
        QuatEuler(0, 75*dt, 0),
        Vec(0, 0, 0),
        GetShapeLocalTransform(ufo.beamFrameShape))

    --Define UFO position
    local newUfoPos = Vec(0, 0, 0)
    --Move UFO if not at destination
    if ufo.flight.timePassed < ufo.flight.time then
        local newFlightPos = progressUfoFlight(dt)
        if newFlightPos == nil then return end
        newUfoPos = VecCopy(newFlightPos)

    --Pre beam
    elseif ufo.beamOn then
        --Make UFO give up if necessary
        if ufo.voxelsTakenByTime <= ufoConsts.voxelsToLeave then
            ufo.giveUpTimer = ufo.giveUpTimer - dt
    
            if ufo.giveUpTimer <= 0 then
                ufoLeave()
                return
            end
        else
            --Reset leave timer
            ufo.giveUpTimer = ufoConsts.timeUntilGiveUp
        end
        
        tickUfoProperties(dt)
        
        --Move ufo up or down based on objects below it, DON'T include offset
        newUfoPos = VecDamp(
            ufo.curPos,
            floatOver(ufo.handle, ufo.curPos, ufoConsts.hoverHeight, ufoConsts.colSize, 1, dt, ufoConsts.staticThreshold),
            0.25,
            dt)
    end
    
    --Apply and add offset and rotate UFO based on time
    local newT = Transform(VecAdd(newUfoPos, randOffset), QuatEuler(0, GetTime()*50, 0))
    SetBodyTransform(ufo.handle, newT)

    --Fire beam
    if ufo.beamOn then
        --Logic for UFO beam
        tickUfoBeam(newT.pos, dt)
    end

    --Play UFO hover sound (louder if beam not on)
    PlayLoop(hoverSnd, newUfoPos, (ufo.beamOn) and 0.25 or 1)
end
--Renders the tractor beam
function renderTractorBeam(start, over)
    --NOTE PARTICLES NOT INCLUDED
    --Beam sprites
    fillVerticallyWithSprites(
        start,
        over, 
        beamSprite, 
        6, 
        3, 
        config.theme.color_r + 0.75, 
        config.theme.color_g + 0.75,
        config.theme.color_b + 0.75,
        (((math.sin(GetTime() * 6) + 1) / 4) + 0.5) * 0.5)
end
function ufoUpdate(dt)
    if ufo.handle == nil then return end
    if not ufo.beamOn then return end

    --Particles
    pullParticles(ufo.beamEnd, ufo.pullSize, ufo.pullPower)
    beamEndParticles(ufo.beamEnd)
    tractorBeamParticles(ufo.beamEnd, ufo.beamDist)
    beamBeginParticles(ufo.offsetPos, ufoConsts.tractorBeamSize, Vec(0, 0.25, 0))
end
--Calculates the end of the UFO tractor beam
function calcBeamEnd(curPos, bodies)
    --Get beam end pos
    QueryRejectBody(ufo.handle)
    for x=1, #bodies do QueryRejectBody(bodies[x]) end
    QueryRequire("large")
    local _, dist = QueryRaycast(curPos,
        Vec(0, -1, 0), ufoConsts.tractorBeamDistance, ufoConsts.tractorBeamSize, false)   
    ufo.beamEnd = VecSub(curPos, Vec(0, dist + ufoConsts.tractorBeamSize, 0))
    ufo.beamDist = dist + ufoConsts.tractorBeamSize
end
--Progress UFO flight
function progressUfoFlight(dt)
    --Progress time
    ufo.flight.timePassed = ufo.flight.timePassed + dt

    if ufo.flight.timePassed >= ufo.flight.time then
        --End reached
        if ufo.flight.destroyOnReachFlightEnd then
            Delete(ufo.handle)
            ufo = {}
            return nil
        end
        beamStateChange(true)    
    end

    --Create new pos
    return
    VecLerp(
        ufo.flight.startingPos,
        ufo.flight.targetPos,
        ufo.flight.easing(ufo.flight.timePassed/ufo.flight.time))
end
--Progress UFO beam
function tickUfoBeam(curPos, dt)
    --Create point light
    PointLight(
        ufo.offsetPos,
        config.theme.color_r,
        config.theme.color_g,
        config.theme.color_b,
        glowPower)

    --Play UFO beam sound
    PlayLoop(beamSnd, curPos, 3)

    --Get every object in the scene
    local bodies = FindBodies("", true)

    --Filter out every moving body in the scene
    local nonMovingBodies = {}
    for x=1, #bodies do
        if not isStationary(bodies[x], ufoConsts.staticThreshold) then
            table.insert(nonMovingBodies, bodies[x])
        end
    end

    --Get beam end pos and store it in ufo.beamEnd
    calcBeamEnd(curPos, nonMovingBodies)
    
    --Render beam
    renderTractorBeam(curPos, ufo.beamEnd)

    --Get objects inside beam pull distance
    QueryRejectBody(ufo.handle)
    QueryRequire("dynamic")
    local foundBodies = QueryAabbBodies(VecSub(ufo.beamEnd,
        Vec(ufo.pullSize, 0, ufo.pullSize)),
        VecAdd(curPos,
            Vec(ufo.pullSize, 0, ufo.pullSize))
    )
    --Filter out too large objects
    local unpullable = {}
    for k, v in ipairs(foundBodies) do
        if IsBodyDynamic(v) then
            local voxels = getBodyVoxelCount(v)

            if ufo.maxAbductVoxels < voxels then
                --It wasn't pullable, add to unpullables...
                table.insert(unpullable, v)
            end
            --It was pullable
        end
    end

    --Get objects in the tractor beam distance
    --Make ray not hit the UFO
    QueryRejectBody(ufo.handle)
    for x=1, #unpullable do
        QueryRejectBody(unpullable[x])
    end
    QueryRequire("dynamic")
    
    local tractorBeamTargetBodies = QueryAabbBodies(VecSub(ufo.beamEnd,
        Vec(ufoConsts.tractorBeamEffectSize, 0, ufoConsts.tractorBeamEffectSize)),
        VecAdd(curPos,
            Vec(ufoConsts.tractorBeamEffectSize, 0, ufoConsts.tractorBeamEffectSize))
    )
    --Hit objects, objects which hit the UFO deletion box
    --Make ray not hit the UFO
    QueryRejectBody(ufo.handle)
    for x=1, #unpullable do
        QueryRejectBody(unpullable[x])
    end
    QueryRequire("dynamic")
    
    local hitShapes = QueryAabbShapes(VecSub(curPos,
        Vec(ufoConsts.colSize, 1, ufoConsts.colSize)),
        VecAdd(curPos,
            Vec(ufoConsts.colSize, 0, ufoConsts.colSize))
    )

    --Apply velocities and abduct objects

    for x=1, #tractorBeamTargetBodies do
        local oldVel = GetBodyVelocity(tractorBeamTargetBodies[x])
        SetBodyVelocity(tractorBeamTargetBodies[x],
            Vec(oldVel[1]/ufoConsts.tractorBeamDamping, 
                oldVel[2] + ufo.tractorBeamPower*dt,
                oldVel[3]/ufoConsts.tractorBeamDamping))
    end
    --Pull all bodies in range
    for x=1, #foundBodies do
        local bodyPos = getBodyPos(foundBodies[x])
        local distanceModifier = getFromPointValue(
            VecDist(Vec(ufo.beamEnd[1], bodyPos[2], ufo.beamEnd[3]), bodyPos),
            ufo.pullSize,
            ufoConsts.distancePowerCurve)

        SetBodyVelocity(foundBodies[x],
            VecAdd(
                GetBodyVelocity(foundBodies[x]), 
                VecScale(
                    Vec(curPos[1] - bodyPos[1], 0, curPos[3] - bodyPos[3]),
                    ufo.pullPower * dt * distanceModifier)))
    end

    --Abduct sound
    if #hitShapes > 0 then
        PlaySound(destroySnd, curPos, 0.5)
    end
    for x=1, #hitShapes do
        local body = GetShapeBody(hitShapes[x])
        local vehicle = GetBodyVehicle(body)

        local voxCount = GetShapeVoxelCount(hitShapes[x])
        ufo.voxelsTakenByTime = ufo.voxelsTakenByTime + voxCount

        dissipateParticles(hitShapes[x])

        --Delete shapes if selection is not a vehicle
        if vehicle + 1 ~= body then      
            local joints = GetShapeJoints(hitShapes[x])
            for x=0, #joints do
                Delete(joints[x])
            end
            Delete(hitShapes[x])      
        end
        --Delete vehicle if vehicle
        if vehicle + 1 == body then
            Delete(vehicle)
        end
    end
    
    --Break static objects
    local breakDir = VecNormalize(
        Vec(math.random(-100, 100), math.random(-100, 100), math.random(-100, 100)))
    local breakStartPos = VecLerp(
        VecLerp(curPos, ufo.beamEnd, math.random(0, 100)/100),
        ufo.beamEnd,
        math.min(math.random(0, ufoConsts.beamDestructionTipBias*100)/100, 1))

    QueryRejectBody(ufo.handle)
    QueryRequire("large")

    local hit, dist = QueryRaycast(
        breakStartPos,
        breakDir,
        ufo.breakDistance,
        0,
        false)   
    if hit then
        local breakPos = VecAdd(VecScale(breakDir, dist), breakStartPos)
        if breakPos[2] >= ufo.beamEnd[2] - 1 then
            MakeHole(breakPos, 2, 1, 0, false)
        end
    end 

    --Break everything in the beam tip region
    local breakPos = 
        VecAdd(
            VecNormalize(
                VecRand(-ufoConsts.tractorBeamSize*2, ufoConsts.tractorBeamSize*2, 2, 1, 0, 1)
            ),
            ufo.beamEnd
        )
    Shoot(breakPos, Vec(0, -1, 0), 'shotgun', 4, dt)
end
--Progress UFO properties
function tickUfoProperties(dt)
    --Progress UFO voxels by time value
    ufo.voxelsTakenByTime = clamp(
        ufo.voxelsTakenByTime - dt*(ufoConsts.voxelHungerGain * ufoConsts.voxelsToMinPower),
        0, ufoConsts.voxelsToMinPower)    
    local powerOfMax = (ufoConsts.voxelsToMinPower - ufo.voxelsTakenByTime)
        / ufoConsts.voxelsToMinPower

    --Apply UFO properties based on voxels by time value
    ufo.pullPower = lerp(4, 10, powerOfMax)
    ufo.tractorBeamPower = lerp(14, 23, powerOfMax)
    ufo.pullSize = lerp(6, 28, powerOfMax)
    ufo.breakDistance = lerp(4, 16, powerOfMax)
    ufo.maxAbductVoxels = lerp(1500, 5000, powerOfMax)
end
--Update UFO offset
function updateUfoPosVariables()
    --Generate offset
    local randOffset = Vec(0, math.sin(GetTime()*ufoConsts.randOffsetSpeed)*ufoConsts.randOffsetPower, 0)

    --Get base pos - last offset
    local bodyPos = getBodyPos(ufo.handle)
    ufo.curPos = VecSub(bodyPos, ufo.lastOffset)
    ufo.offsetPos = bodyPos

    return randOffset
end